/*  engine.c - Parabix XML parsing engine.
    Copyright (c) 2007, 2008, Robert D. Cameron and Dan Lin.
    Licensed to the public under the Open Software License 3.0.
    Licensed to International Characters, Inc., under the Academic
    Free License 3.0.
*/

#include "engine.h"
#include "byteplex.h"
#include "xmldecl.h"
#include "bytelex.h"
#include "bitlex.h"
#include "contentmodel.h"
#include "contentmodel.c"
#include "xml_error.h"

#include <assert.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <string>
#include <iostream>
using namespace std;

inline char * copy_string (unsigned char * s, int lgth){
	char * d = new char[lgth+1];
	memcpy(d, (char *)s,lgth);
	d[lgth] = '\0';
	return d;
}

inline char * cat_string (char * s1, char * s2, int lgth1, int lgth2){
	char * s = new char[lgth1 + lgth2 + 1];
	memcpy(s, s1,lgth1);
	memcpy(&s[lgth1],s2,lgth2);
	s[lgth1 + lgth2] = '\0';
	return s;
}


template <WorkingCharacterSet W>
Parser_Interface<W> * Parser_Interface<W>::ParserFactory(const char * filename) {

	int chars_read;
	unsigned char signature[4];
	FILE * infile;
	infile = fopen(filename, "rb");
  	if (!infile) {
		fprintf(stderr, "Error: cannot open %s for input.\n", filename);
		exit(-1);
	}
	fread(signature,1,4,infile);
	Entity_Info * e = new Entity_Info;
	Model_Info * m = new Model_Info;
	e->AnalyzeSignature(signature);
	Byteplex * b = Byteplex::ByteplexFactory(e, infile);
	b->InitializeBuffer(signature,4);
	b->DoByteplex();
	b->PreparePseudoASCII_Stream();

	if (e->code_unit_base == ASCII) {
		XML_Decl_Parser<ASCII> decl_parser(b);
		decl_parser.ReadXMLInfo(*e);
		if (e->code_unit_size == SingleByte) {
			if (!(e->has_encoding_decl) || at_UTF_8(e->encoding))
				return new ParsingEngine< UTF8_Buffer, W>(e, m, b, false);
			else return new ParsingEngine< X8_Buffer<ASCII>, W>(e, m, b, false);
		}
		else if (e->code_unit_size == DoubleByte) {
			return new ParsingEngine<U16_Buffer, W>(e, m, b, false);
		}
		else if (e->code_unit_size == QuadByte) {
			return new ParsingEngine<U32_Buffer, W>(e, m, b, false);
		}
	}
	else /* if (e->code_unit_base == EBCDIC) */ {
		XML_Decl_Parser<EBCDIC> decl_parser(b);
		decl_parser.ReadXMLInfo(*e);
		return new ParsingEngine< X8_Buffer<EBCDIC>, W>(e, m, b, false);
	}
}

template <WorkingCharacterSet W>
Parser_Interface<W> * Parser_Interface<W>::ParserFactory(const char * filename, Model_Info * m) {

	int chars_read;
	unsigned char signature[4];
	FILE * infile;
	infile = fopen(filename, "rb");
  	if (!infile) {
		fprintf(stderr, "Error: cannot open %s for input.\n", filename);
		exit(-1);
	}
	fread(signature,1,4,infile);
	Entity_Info * e = new Entity_Info;
	e->AnalyzeSignature(signature);
	Byteplex * b = Byteplex::ByteplexFactory(e, infile);
	b->InitializeBuffer(signature,4);
	b->DoByteplex();
	b->PreparePseudoASCII_Stream();
	if (e->code_unit_base == ASCII) {
		XML_Decl_Parser<ASCII> decl_parser(b);
		decl_parser.ReadXMLInfo(*e);
		if (e->code_unit_size == SingleByte) {
			return new ParsingEngine< X8_Buffer<ASCII>, W>(e, m, b, true);
		}
		else if (e->code_unit_size == DoubleByte) {
			return new ParsingEngine<U16_Buffer, W>(e, m, b, true);
		}
		else if (e->code_unit_size == QuadByte) {
			return new ParsingEngine<U32_Buffer, W>(e, m, b, true);
		}
	}
	else /* if (e->code_unit_base == EBCDIC) */ {
		XML_Decl_Parser<EBCDIC> decl_parser(b);
		decl_parser.ReadXMLInfo(*e);
		return new ParsingEngine< X8_Buffer<EBCDIC>, W>(e, m, b, true);
	}
}

template <WorkingCharacterSet W>
Parser_Interface<W> * Parser_Interface<W>::ParserFactory(const char * byte_buffer, int byte_count, Entity_Info * e1, Model_Info * m){
	Entity_Info * e = new Entity_Info;
	e->BOM_units = 0;
	e->code_unit_base=e1->code_unit_base;
	e->code_unit_size=e1->code_unit_size;
	e->version=e1->version;
	e->encoding=e1->encoding;
	e->content_start = 0;
	Byteplex * b = Byteplex::ByteplexFactory(e, (unsigned char *) byte_buffer, byte_count);
	b->DoByteplex();
	b->PreparePseudoASCII_Stream();
	if (e->code_unit_base == ASCII) {
		if (e->code_unit_size == SingleByte) {
			return new ParsingEngine< X8_Buffer<ASCII>, W>(e, m, b, false);
		}
		else if (e->code_unit_size == DoubleByte) {
			return new ParsingEngine<U16_Buffer, W>(e, m, b, false);
		}
		else if (e->code_unit_size == QuadByte) {
			return new ParsingEngine<U32_Buffer, W>(e, m, b, false);
		}
	}
	else /* if (e->code_unit_base == EBCDIC) */ {
		return new ParsingEngine< X8_Buffer<EBCDIC>, W>(e, m, b, false);
	}
}

template <WorkingCharacterSet W>
Parser_Interface<W>::~Parser_Interface() {
}


template <WorkingCharacterSet W>
bool Parser_Interface<W>::has_ByteOrderMark() {
	return entity_Info->BOM_units > 0;
}

template <WorkingCharacterSet W>
XML_version Parser_Interface<W>::get_version() {
	return entity_Info->version;
}

template <WorkingCharacterSet W>
XML_standalone Parser_Interface<W>::standalone_status() {
	return entity_Info->standalone;
}

template <WorkingCharacterSet W>
bool Parser_Interface<W>::has_EncodingDecl() {
	return entity_Info->has_encoding_decl;
}

template <WorkingCharacterSet W>
unsigned char * Parser_Interface<W>::get_Encoding() {
	return entity_Info->encoding;
}

template <class B, WorkingCharacterSet W>
inline unsigned char * ParsingEngine<B, W>::GetCodeUnitPtr(int pos) {
	int rel_pos = pos - buffer_base_pos;
	return &((unsigned char *) (byteplex->src_buffer))[rel_pos * (int) B::Size];
}

template <>
inline unsigned char * ParsingEngine<UTF8_Buffer, UTF_8>::GetCodeUnitPtr(int pos) {
	int rel_pos = pos - buffer_base_pos;
	return &((unsigned char *) (x8data))[rel_pos];
}




template <class B, WorkingCharacterSet W>
ParsingEngine<B, W>::ParsingEngine(Entity_Info * e, Model_Info * m, Byteplex * b, bool is_external) : Parser_Interface<W> () {
	Parser_Interface<W>::entity_Info = e;
	Parser_Interface<W>::model_info = m;
	byteplex = b;

//	m->symbol_table = new Symbol_Table();
//	m->SimpleEntity("lt", "<");
//	m->SimpleEntity("gt", ">");
//	m->SimpleEntity("amp", "&");
//	m->SimpleEntity("quot", "\"");
//	m->SimpleEntity("apos", "'");
	m->symbol_table->version = e->version;

	StrictWellFormedness=false;
	LastAttOccurrence.assign(m->globalAttributeCount+1, 0);


	bitplex = new Bitplex;
	buf = (LexicalStreamSet *) simd_new(sizeof(LexicalStreamSet)/PACKSIZE);

  /* Install sentinels for every lexical item stream*/
#ifdef TEMPLATED_SIMD_LIB
	BitBlock sentinel_value = simd<1>::constant<1>();
#endif
#ifndef TEMPLATED_SIMD_LIB
	BitBlock sentinel_value = simd_const_1(1);
#endif

#ifdef OPTIMIZE_SHORT_SCAN
	sentinel_value = sisd_sfli(sentinel_value, 8*sizeof(unsigned long));
#endif

	for (int j = minLexicalItem; j < LexicalItemCount; j++) {
		buf->item_stream[j][BUFFER_BLOCKS] = sentinel_value;
	}

	buffer_base_pos = 0;
	buffer_rel_pos = e->content_start;
	buffer_limit_pos = min(BUFFER_SIZE, byteplex->units_in_buffer);
	int blocks_in_buffer = (buffer_limit_pos + BLOCKSIZE - 1)/BLOCKSIZE;
	x8data = byteplex->x8data;
	lexer = Lexer<B::Base>::LexerFactory(e, buf);
	bitplex->TransposeToBitStreams(byteplex->x8data, blocks_in_buffer);
	lexer->AnalyzeBuffer(bitplex->x8basis, buffer_base_pos, buffer_rel_pos, buffer_limit_pos);
}

template <class B, WorkingCharacterSet W>
ParsingEngine<B, W>::~ParsingEngine() {
  // How do we do this?  Parser_Interface<W>::model_info->~Model_Info();
  Parser_Interface<W>::entity_Info->~Entity_Info();
  delete byteplex;
  delete bitplex;
  simd_delete((SIMD_type *) buf);
  delete lexer;
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::AdvanceBuffers(){
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == ADVANCE_BUFFERS)
	code_clocker->start_interval();
#endif

	int advance_amt = text_or_markup_start - buffer_base_pos;
	advance_amt &= -PACKSIZE; // maintain alignment
	byteplex->AdvanceInputBuffer(advance_amt);
	buffer_base_pos += advance_amt;
	buffer_rel_pos -= advance_amt;
	buffer_limit_pos = min(BUFFER_SIZE, byteplex->units_in_buffer);
	int blocks_in_buffer = (buffer_limit_pos + BLOCKSIZE - 1)/BLOCKSIZE;
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BYTEPLEX)
	code_clocker->start_interval();
#endif
	byteplex->DoByteplex();
	byteplex->PreparePseudoASCII_Stream();
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BYTEPLEX)
	code_clocker->end_interval(buffer_limit_pos);
#endif
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BITPLEX)
	code_clocker->start_interval();
#endif
	bitplex->TransposeToBitStreams(byteplex->x8data, blocks_in_buffer);
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BITPLEX)
	code_clocker->end_interval(buffer_limit_pos);
#endif
   	lexer->AnalyzeBuffer(bitplex->x8basis, buffer_base_pos, buffer_rel_pos, buffer_limit_pos);
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == ADVANCE_BUFFERS)
	code_clocker->end_interval(buffer_limit_pos);
#endif

}

template <>
void ParsingEngine<U16_Buffer, UTF_16>::AdvanceBuffers(){
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == ADVANCE_BUFFERS)
	code_clocker->start_interval();
#endif

	int advance_amt = text_or_markup_start - buffer_base_pos;
	advance_amt &= -PACKSIZE; // maintain alignment
	byteplex->AdvanceInputBuffer(advance_amt);
	buffer_base_pos += advance_amt;
	buffer_rel_pos -= advance_amt;
	buffer_limit_pos = min(BUFFER_SIZE, byteplex->units_in_buffer);
	int blocks_in_buffer = (buffer_limit_pos + BLOCKSIZE - 1)/BLOCKSIZE;
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BYTEPLEX)
	code_clocker->start_interval();
#endif
	byteplex->DoByteplex();
	if (at_UTF_16(Parser_Interface<UTF_16>::entity_Info->encoding)) ((U16_Buffer *) byteplex)->Validate_UTF16();
	byteplex->PreparePseudoASCII_Stream();
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BYTEPLEX)
	code_clocker->end_interval(buffer_limit_pos);
#endif
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BITPLEX)
	code_clocker->start_interval();
#endif
	bitplex->TransposeToBitStreams(byteplex->x8data, blocks_in_buffer);
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == BITPLEX)
	code_clocker->end_interval(buffer_limit_pos);
#endif
   	lexer->AnalyzeBuffer(bitplex->x8basis, buffer_base_pos, buffer_rel_pos, buffer_limit_pos);
#if defined(PAPI) and defined(CODE_CLOCKING) and (CODE_CLOCKING == ADVANCE_BUFFERS)
	code_clocker->end_interval(buffer_limit_pos);
#endif

}

template <class B, WorkingCharacterSet W>
inline unsigned char * ParsingEngine<B, W>::cur() const {
  return &((unsigned char *) x8data)[buffer_rel_pos];
}

template <class B, WorkingCharacterSet W>
inline int ParsingEngine<B, W>::AbsPos() const {
  return buffer_base_pos + buffer_rel_pos;
}

template <class B, WorkingCharacterSet W>
inline int ParsingEngine<B, W>::LengthFrom(int start_pos) const {
  return buffer_base_pos + buffer_rel_pos - start_pos;
}



template <class B, WorkingCharacterSet W>
inline int ParsingEngine<B, W>::BufferRelPos() const {
  return buffer_rel_pos;
}


template <class B, WorkingCharacterSet W>
inline bool ParsingEngine<B, W>::at_EOF() const {
  return (buffer_rel_pos >= buffer_limit_pos) &&
         (buffer_limit_pos < BUFFER_SIZE);
}

//template <class B, WorkingCharacterSet W>
//inline void ParsingEngine<B, W>::Advance(int n) {
//	buffer_rel_pos += n;
//  if (buffer_rel_pos >= BUFFER_SIZE) {
//	Parser_Interface<W>::FinalizeBuffer_action();
//	AdvanceBuffers();
//  }
//}

#define Advance(n) \
do {\
	buffer_rel_pos += n; \
	if (buffer_rel_pos >= BUFFER_SIZE) {	\
		Parser_Interface<W>::FinalizeBuffer_action();\
	AdvanceBuffers();\
	}\
} while(0)


template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::AdjustBufferEndForIncompleteSequences() {
}

template <>
void ParsingEngine<UTF8_Buffer, UTF_8>::AdjustBufferEndForIncompleteSequences() {
	if (*(cur()-1) >= 0xC0) buffer_rel_pos--;
	else if (*(cur()-2) >= 0xE0) buffer_rel_pos -= 2;
	else if (*(cur()-3) >= 0xF0) buffer_rel_pos -= 3;
}

template <>
void ParsingEngine<U16_Buffer, UTF_8>::AdjustBufferEndForIncompleteSequences() {
	unsigned short last_u16_unit = *(GetCodeUnitPtr(AbsPos()-1));
	if ((last_u16_unit >= 0xD800) & (last_u16_unit <= 0xDC00)) buffer_rel_pos--;
}

template <>
void ParsingEngine<UTF8_Buffer, UTF_16>::AdjustBufferEndForIncompleteSequences() {
	if (*(cur()-1) >= 0xC0) buffer_rel_pos--;
	else if (*(cur()-2) >= 0xE0) buffer_rel_pos -= 2;
	else if (*(cur()-3) >= 0xF0) buffer_rel_pos -= 3;
}

template <>
void ParsingEngine<U16_Buffer, UTF_16>::AdjustBufferEndForIncompleteSequences() {
	unsigned short last_u16_unit = *(GetCodeUnitPtr(AbsPos()-1));
	if ((last_u16_unit >= 0xD800) & (last_u16_unit <= 0xDC00)) buffer_rel_pos--;
}



#ifdef OPTIMIZE_SHORT_SCAN
//
//  Inline ScanTo with unrolled first test that should almost always
//  succeed for short scans.
#define ScanTo(item) \
do {\
	unsigned long segment = bitstream_segment_from(buf->item_stream[item], buffer_rel_pos);\
	if (segment != 0) buffer_rel_pos += cfzl(segment);\
	else {\
		buffer_rel_pos = (buffer_rel_pos & -8) + 8*sizeof(unsigned long);\
		buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);\
		while (buffer_rel_pos >= BUFFER_SIZE) {\
			buffer_rel_pos = BUFFER_SIZE;\
			AdjustBufferEndForIncompleteSequences();\
			Parser_Interface<W>::FinalizeBuffer_action();\
			AdvanceBuffers();\
			buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);\
		}\
	}\
} while(0)

// The following version seems cleaner, but measured mispredictions are higher
// #define ScanTo(item) \
// do {\
// 	unsigned long segment = bitstream_segment_from(buf->item_stream[item], buffer_rel_pos);\
// 	while (unlikely (segment == 0)) {\
// 		buffer_rel_pos = (buffer_rel_pos & -8) + 8*sizeof(unsigned long);\
// 		if (buffer_rel_pos >= BUFFER_SIZE) {\
// 			buffer_rel_pos = BUFFER_SIZE;\
// 			AdjustBufferEndForIncompleteSequences();\
// 			Parser_Interface<W>::FinalizeBuffer_action();\
// 			AdvanceBuffers();\
// 		}\
// 		segment = bitstream_segment_from(buf->item_stream[item], buffer_rel_pos);\
// 	}\
// 	buffer_rel_pos += cfzl(segment);\
// } while(0)
//
// #define ScanTextTo(item) \
// do {\
// 	unsigned long segment = bitstream_segment_from(buf->item_stream[item], buffer_rel_pos);\
// 	text_or_markup_start = AbsPos();\
// 	if (segment != 0) buffer_rel_pos += cfzl(segment);\
// 	else {\
// 		buffer_rel_pos = (buffer_rel_pos & -8) + 8*sizeof(unsigned long);\
// 		buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);\
// 		while (buffer_rel_pos >= BUFFER_SIZE) {\
// 			buffer_rel_pos = BUFFER_SIZE;\
// 			AdjustBufferEndForIncompleteSequences();\
// 			Text_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start), true);\
// 			text_or_markup_start = AbsPos();\
// 			Parser_Interface<W>::FinalizeBuffer_action();\
// 			AdvanceBuffers();\
// 			buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);\
// 		}\
// 	}\
// } while(0)

template <class B, WorkingCharacterSet W>
inline void ParsingEngine<B, W>::ScanTextTo(int item) {
	text_or_markup_start = AbsPos();
	unsigned long segment = bitstream_segment_from(buf->item_stream[item], buffer_rel_pos);
	if (segment != 0) buffer_rel_pos += cfzl(segment);
	else {
		buffer_rel_pos = (buffer_rel_pos & -8) + 8*sizeof(unsigned long);
		buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);
		while (buffer_rel_pos >= BUFFER_SIZE) {
			buffer_rel_pos = BUFFER_SIZE;
			AdjustBufferEndForIncompleteSequences();
			Text_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start), true);
			text_or_markup_start = AbsPos();
			Parser_Interface<W>::FinalizeBuffer_action();
			AdvanceBuffers();
			buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);
		}
	}
}

#endif

#ifndef OPTIMIZE_SHORT_SCAN

// #define ScanTo(item) \
// do {\
//   buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);\
//   while (buffer_rel_pos >= BUFFER_SIZE) {\
// 	AdjustBufferEndForIncompleteSequences();\
// 	Parser_Interface<W>::FinalizeBuffer_action();\
// 	AdvanceBuffers();\
// 	buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);\
//   }\
// } while(0)


template <class B, WorkingCharacterSet W>
inline void ParsingEngine<B, W>::ScanTo(int item) {
	buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);
	while (buffer_rel_pos >= BUFFER_SIZE) {
		AdjustBufferEndForIncompleteSequences();
		Parser_Interface<W>::FinalizeBuffer_action();
		AdvanceBuffers();
		buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);
 	}
}

template <class B, WorkingCharacterSet W>
inline void ParsingEngine<B, W>::ScanTextTo(int item) {
	text_or_markup_start = AbsPos();
	buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);
	while (buffer_rel_pos >= BUFFER_SIZE) {
		AdjustBufferEndForIncompleteSequences();
		Text_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start), true);
		text_or_markup_start = AbsPos();
		Parser_Interface<W>::FinalizeBuffer_action();
		AdvanceBuffers();
		buffer_rel_pos = bitstream_scan(buf->item_stream[item], buffer_rel_pos);
	}
}
#endif

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::WF_Error (XML_Constraint errCode) {
	printf("Error at position %i in input.\n", AbsPos());
	ShowConstraintError(errCode);
	exit(-1);
//	Error_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
}


template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Validity_Error (XML_Constraint errCode) {
	printf("Error at position %i in input.\n", AbsPos());
	ShowConstraintError(errCode);
	exit(-1);
//	Error_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Syntax_Error (XML_NonTerminal errNT) {
	printf("Error at position %i in input.\n", AbsPos());
	ShowSyntaxError(errNT);
	exit(-1);
//	Error_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
}


/* Parse a comment beginning "<!--" */
template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_Comment() {

	Advance(4); /* Skip "<!--". */
	ScanTo(Hyphen);
	while (!at_DoubleHyphen<B::Base>(cur())) {
		if(at_EOF())
			Syntax_Error(NT_CDSect);
		Advance(2); /* Skip hyphen-nonhyphen pair */
		ScanTo(Hyphen);
	}
	if (at_Comment_End<B::Base>(cur())) {
		Advance(3); /* Skip "-->". */
		Comment_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
	}
	else {
		Advance(2);  /* "--" */
		Syntax_Error(NT_Comment);
	}
}

/* Parse an end tag beginning "</" */
template <class B, WorkingCharacterSet W>
inline void ParsingEngine<B, W>::Parse_EndTag() {
	Advance(2); /* Skip "</". */
	int nameID = Parse_Name();
	if (AtChar<B::Base,'>'>(cur())) {
		Advance(1);
		EndTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
	}
        else {
		ScanTo(NonWS);
		if (AtChar<B::Base,'>'>(cur())) {
			Advance(1);
			EndTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else Syntax_Error(NT_ETag);
        }
}

/* Parse a CDATA section beginning "<![CDATA". */
template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_CDATA() {
		Advance(8); /* Skip "<![CDATA". */
        if (!AtChar<B::Base,'['>(cur())) {
 		Syntax_Error(NT_CDStart);
        }
        else {
		Advance(1);
		CDATA_start_action(GetCodeUnitPtr(text_or_markup_start));
		text_or_markup_start = AbsPos();
		ScanTextTo(CD_End_check);
		while (!at_CDATA_End<B::Base>(cur())) {
			if (at_EOF())
				Syntax_Error(NT_CDSect);
			Advance(1);
			ScanTextTo(CD_End_check);
		}
		Text_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start), true);
		Advance(3); /* Skip "]]>". */
		CDATA_end_action(GetCodeUnitPtr(AbsPos()));
	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_EntityRef() {
    Advance(1);  // skip "&"
	int nameID = Parse_Name();  /* Name delimiter */
    if (!AtChar<B::Base,';'>(cur())) {
 		Syntax_Error(NT_Reference);
    }
	else {
		Advance(1);
		Reference_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));

		//	The following code will replace Reference_Action.
		GEntity_info * this_info;
		Parser_Interface<W> * entity_parser;
		int entityID = Parser_Interface<W>::model_info->GlobalGEntityTable[nameID];
		if (entityID == 0)
			WF_Error(wfErr_wf_entdeclared);
		else{
			this_info = Parser_Interface<W>::model_info->GEntityData[entityID-1];
			if (this_info->is_external){

	     		if (Parser_Interface<W>::entity_Info->standalone != Standalone_no)
	     			WF_Error(wfErr_NoExternalRefs);
	     		else {
					entity_parser = ParserFactory(this_info->systemLiteral, Parser_Interface<W>::model_info);
					entity_parser->Parse_WF_Content();
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
	     		}
			}
			else {
				if (this_info->is_simple == true);
//					printf("Entity is %s\n",this_info->ReplacementText);
				else{
//					printf("Not a simple text: %s\n",this_info->ReplacementText);
					entity_parser = ParserFactory(this_info->ReplacementText, strlen(this_info->ReplacementText),Parser_Interface<W>::entity_Info, Parser_Interface<W>::model_info);
					entity_parser->Parse_WF_Content();
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
				}
			}
		}

	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_EntityRef_inMixed(symbol_set_t elems) {
    Advance(1);  // skip "&"
	int nameID = Parse_Name();  /* Name delimiter */
    if (!AtChar<B::Base,';'>(cur())) {
 		Syntax_Error(NT_Reference);
    }
	else {
		Advance(1);
		Reference_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));

		//	The following code will replace Reference_Action.
		GEntity_info * this_info;
		Parser_Interface<W> * entity_parser;
		int entityID = Parser_Interface<W>::model_info->GlobalGEntityTable[nameID];
		if (entityID == 0)
			WF_Error(wfErr_wf_entdeclared);
		else{
			this_info = Parser_Interface<W>::model_info->GEntityData[entityID-1];
			if (this_info->is_external){

	     		if (Parser_Interface<W>::entity_Info->standalone != Standalone_no)
	     			WF_Error(wfErr_NoExternalRefs);
	     		else {
					entity_parser = ParserFactory(this_info->systemLiteral, Parser_Interface<W>::model_info);
					entity_parser->Parse_MixedContent(elems);
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
	     		}
			}
			else {
				if (this_info->is_simple == true);
//					printf("Entity is %s\n",this_info->ReplacementText);
				else{
//					printf("Not a simple text: %s\n",this_info->ReplacementText);
					entity_parser = ParserFactory(this_info->ReplacementText, strlen(this_info->ReplacementText),Parser_Interface<W>::entity_Info, Parser_Interface<W>::model_info);
					entity_parser->Parse_MixedContent(elems);
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
				}
			}
		}

	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_EntityRef_inAnyContent() {
    Advance(1);  // skip "&"
	int nameID = Parse_Name();  /* Name delimiter */
    if (!AtChar<B::Base,';'>(cur())) {
 		Syntax_Error(NT_Reference);
    }
	else {
		Advance(1);
		Reference_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));

		//	The following code will replace Reference_Action.
		GEntity_info * this_info;
		Parser_Interface<W> * entity_parser;
		int entityID = Parser_Interface<W>::model_info->GlobalGEntityTable[nameID];
		if (entityID == 0)
			WF_Error(wfErr_wf_entdeclared);
		else{
			this_info = Parser_Interface<W>::model_info->GEntityData[entityID-1];
			if (this_info->is_external){

	     		if (Parser_Interface<W>::entity_Info->standalone != Standalone_no)
	     			WF_Error(wfErr_NoExternalRefs);
	     		else {
					entity_parser = ParserFactory(this_info->systemLiteral, Parser_Interface<W>::model_info);
					entity_parser->Parse_AnyContent();
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
	     		}
			}
			else {
				if (this_info->is_simple == true);
//					printf("Entity is %s\n",this_info->ReplacementText);
				else{
//					printf("Not a simple text: %s\n",this_info->ReplacementText);
					entity_parser = ParserFactory(this_info->ReplacementText, strlen(this_info->ReplacementText),Parser_Interface<W>::entity_Info, Parser_Interface<W>::model_info);
					entity_parser->Parse_AnyContent();
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
				}
			}
		}

	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_ValidEntityRef(CM_RegExp * cre, int & cur_state) {
    Advance(1);  // skip "&"
	int nameID = Parse_Name();  /* Name delimiter */
    if (!AtChar<B::Base,';'>(cur())) {
 		Syntax_Error(NT_Reference);
    }
	else {
		Advance(1);
		Reference_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));

		//	The following code will replace Reference_Action.
		GEntity_info * this_info;
		Parser_Interface<W> * entity_parser;
		int entityID = Parser_Interface<W>::model_info->GlobalGEntityTable[nameID];
		if (entityID == 0)
			WF_Error(wfErr_wf_entdeclared);
		else{
			this_info = Parser_Interface<W>::model_info->GEntityData[entityID-1];
			if (this_info->is_external){

	     		if (Parser_Interface<W>::entity_Info->standalone != Standalone_no)
	     			WF_Error(wfErr_NoExternalRefs);
	     		else {
					entity_parser = ParserFactory(this_info->systemLiteral, Parser_Interface<W>::model_info);
					entity_parser->Parse_ValidContent(cre, cur_state);
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
	     		}
			}
			else {
				if (this_info->is_simple == true);
//					printf("Entity is %s\n",this_info->ReplacementText);
				else{
//					printf("Not a simple text: %s\n",this_info->ReplacementText);
					entity_parser = ParserFactory(this_info->ReplacementText, strlen(this_info->ReplacementText),Parser_Interface<W>::entity_Info, Parser_Interface<W>::model_info);
					entity_parser->Parse_ValidContent(cre, cur_state);
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
				}
			}
		}

	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_CharRef() {
        Advance(2);  // skip "&#"
        int ch_val = 0;
        if (AtChar<B::Base,'x'>(cur())) {
        	Advance(1);
        	while(at_HexDigit<B::Base>(cur())){
        		ch_val = HexVal<B::Base>(cur()[0]) + (ch_val<<4);
        		if (ch_val> 0x10FFFF )
        			WF_Error(wfErr_wf_Legalchar);
        		Advance(1);
        	}
        }
        else {
        	while(at_Digit<B::Base>(cur())){
        		ch_val = DigitVal<B::Base>(cur()[0]) + ch_val*10;
        		if (ch_val> 0x10FFFF )
        			WF_Error(wfErr_wf_Legalchar);
        		Advance(1);
        	}
        }
        if ((ch_val == 0x0) || ((ch_val | 0x7FF) == 0xDFFF)|| ((ch_val | 0x1) == 0xFFFF))
				WF_Error(wfErr_wf_Legalchar);
		else  if (Parser_Interface<W>::entity_Info->version != XML_1_1)
			if (((ch_val < 0x20) && (ch_val != 0x9) && (ch_val != 0xD) && (ch_val != 0xA)))
				WF_Error(wfErr_wf_Legalchar);

        if (!AtChar<B::Base,';'>(cur())) {
 			Syntax_Error(NT_CharRef);
        }
		else {
			Advance(1);
			Reference_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_PI (){
	int nameID;
	Advance(2); /* Skip "<?". */
	int target_start = AbsPos();
	if (at_XxMmLll<B::Base>(cur())) {
		nameID = Parse_Name();
		if (AbsPos() - target_start == 3) Syntax_Error(NT_PI);
	}
	else nameID = Parse_Name();
	PI_Target_action(GetCodeUnitPtr(target_start), LengthFrom(target_start));
	if (!at_PI_End<B::Base>(cur())) requireWS();
	ScanTo(QMark);
	while (!at_PI_End<B::Base>(cur())) {
		if(at_EOF())
			Syntax_Error(NT_PI);
		Advance(1);
		ScanTo(QMark);
	}
	Advance(2); /* Skip "?>". */
	PI_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
}

/* Parse a start or empty element tag. */
template <class B, WorkingCharacterSet W>
inline void ParsingEngine<B, W>::Parse_StartTag (){
	int att_name_start;
	int att_val_start;
	int att_name_end, att_val_end;
	unsigned char quoteCh;
	Advance(1);
	int nameID = Parse_Name();  /* Name delimiter: WS, "/" or ">" */
	ElementName_action(GetCodeUnitPtr(text_or_markup_start+1), LengthFrom(text_or_markup_start+1));
	/* The following test optimizes the most common case of a
	start tag with no attributes.  */
	if (AtChar<B::Base,'>'>(cur())) {
		Advance(1);
		StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
	}
	else {
		ScanTo(NonWS);
		if (AtChar<B::Base,'>'>(cur())) {
			Advance(1);
			StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else if (at_EmptyElementDelim<B::Base>(cur())) {
			Advance(2);
			EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else do {
			/* Must be an attribute-value pair or error. */
                        att_name_start = AbsPos();
			int att_nameID = Parse_Name();
                        att_name_end = AbsPos();

			int attID = Parser_Interface<W>::model_info->getOrInsertGlobalAttName(att_nameID);
 			if (attID >= LastAttOccurrence.size()) LastAttOccurrence.push_back(0);
 			else {
 				if (LastAttOccurrence[attID] > text_or_markup_start) {
					WF_Error(wfErr_uniqattspec); /* Duplicate attribute. */
					break;
 				}
 			}
 		 	LastAttOccurrence[attID] = att_name_start;
			/* The following optimized tests handle the frequently occurring
			case that there are no blanks on either side of the equals sign.
			In many cases, the very first test handles 100% of actual
			attribute-value pairs encountered. */
			if (at_EqualsQuote<B::Base>(cur())) Advance(1);
			else {
				ScanTo(NonWS);
				if (!AtChar<B::Base,'='>(cur())) {
					Syntax_Error(NT_STag);
					break;
				}
				Advance(1);
				ScanTo(NonWS);
				if (!AtQuote<B::Base>(cur())) {
					Syntax_Error(NT_STag);
					break;
				}
			}
			att_val_start = AbsPos()+1;
			Parse_AttValue();
			att_val_end = AbsPos()-1;
			if (at_xmlns<B::Base>(cur()+att_name_start-AbsPos())) {
				Namespace_action(GetCodeUnitPtr(att_name_start), att_name_end - att_name_start,
						 GetCodeUnitPtr(att_val_start), att_val_end - att_val_start);
			}
			else {
				AttributeValue_action(GetCodeUnitPtr(att_name_start), att_name_end - att_name_start,
						 GetCodeUnitPtr(att_val_start), att_val_end - att_val_start);
			}
			/* Now check for end or repeat. Avoid whitespace scan if possible.*/
			if (AtChar<B::Base,'>'>(cur())) {
				Advance(1);
				StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (at_EmptyElementDelim<B::Base>(cur())) {
				Advance(2);
				EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			ScanTo(NonWS);
			if (AtChar<B::Base,'>'>(cur())) {
				Advance(1);
				StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (at_EmptyElementDelim<B::Base>(cur())) {
				Advance(2);
				EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (AbsPos() == att_val_end + 1) {
				/* No WS following att value */
				Syntax_Error(NT_STag);
				break;
			}
		} while (1);
	}
}

template <class B, WorkingCharacterSet W>
inline void ParsingEngine<B, W>::text_if_nonnull_action(bool more){
	if (AbsPos() > text_or_markup_start) {
		Text_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start), more);
		text_or_markup_start = AbsPos();
	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_WF_EndTag(int nameID) {
	Advance(2);
	int end_nameID = Parse_Name();
	if(end_nameID != nameID)
		WF_Error(wfErr_GIMatch);
	if (AtChar<B::Base,'>'>(cur())) {
		Advance(1);
		Parser_Interface<W>::EndTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
	}
    else {
		ScanTo(NonWS);
		if (AtChar<B::Base,'>'>(cur())) {
			Advance(1);
			Parser_Interface<W>::EndTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else Syntax_Error(NT_ETag);
    }
}

// template <>
// void ParsingEngine<UTF8_Buffer, UTF_8>::Parse_WF_EndTag(int nameID) {
// 	Advance(2); /* Skip "</". */
//
// 	int name_start = AbsPos();
// //	ScanTo(NameFollow);
// //	int lgth = AbsPos()-name_start;
//
// #if (not defined(OMISSION)) or ((OMISSION != END_TAG_MATCHING)  and (OMISSION != NAME_LOOKUP))
// 	char * start_elem_name = Parser_Interface<UTF_8>::model_info->symbol_table->Get_UTF8_name(nameID);
// 	int lgth = Parser_Interface<UTF_8>::model_info->symbol_table->Get_UTF8_lgth(nameID);
// 	char * end_elem_name = &((char *) x8data)[buffer_rel_pos];
//
// #ifdef TEMPLATED_SIMD_LIB
// 	BytePack byte_compare =  simd<8>::eq(sisd_load_unaligned((BytePack *) end_elem_name),
//              						   sisd_load_unaligned((BytePack *) start_elem_name));
// #endif
// #ifndef TEMPLATED_SIMD_LIB
// 	BytePack byte_compare =  simd_eq_8(sisd_load_unaligned((BytePack *) end_elem_name),
//              						   sisd_load_unaligned((BytePack *) start_elem_name));
// #endif
// 	if (lgth < 16) {
// 		int expected_bits = ~(-1 << lgth);
// 	    if ((_mm_movemask_epi8(byte_compare) & expected_bits) != expected_bits) {
// 	         	WF_Error(wfErr_GIMatch);
// 	    }
// 	}
// 	else {
// 	    /* Must compare with bytes beyond the first 16.  Set up to
// 	       compare 16 bytes at a time, with the first additional compare
// 	       overlapping with the first byte_compare. */
// 	    int pos = (lgth - 1) % PACKSIZE + 1;
// #ifdef TEMPLATED_SIMD_LIB
// 	    byte_compare =  simd_or(byte_compare, simd<8>::eq(sisd_load_unaligned((BytePack *) &end_elem_name[pos]),
// 	                        							sisd_load_unaligned((BytePack *) &start_elem_name[pos])));
// #endif
// #ifndef TEMPLATED_SIMD_LIB
// 	    byte_compare =  simd_or(byte_compare, simd_eq_8(sisd_load_unaligned((BytePack *) &end_elem_name[pos]),
// 	                        							sisd_load_unaligned((BytePack *) &start_elem_name[pos])));
// #endif
// 	    pos += 16;
// 	    while (pos < lgth) {
// 	    	if (_mm_movemask_epi8(byte_compare) != 0xFFFF) {
// 	        	WF_Error(wfErr_GIMatch);
// 	        }
// #ifdef TEMPLATED_SIMD_LIB
// 	        byte_compare =  simd<8>::eq(sisd_load_unaligned((BytePack *) &end_elem_name[pos]),
// 	                        		  sisd_load_unaligned((BytePack *) &start_elem_name[pos]));
// #endif
// #ifndef TEMPLATED_SIMD_LIB
// 	        byte_compare =  simd_eq_8(sisd_load_unaligned((BytePack *) &end_elem_name[pos]),
// 	                        		  sisd_load_unaligned((BytePack *) &start_elem_name[pos]));
// #endif
// 	        pos += 16;
// 	    }
// 	    if (_mm_movemask_epi8(byte_compare) != 0xFFFF) {
// 	        	WF_Error(wfErr_GIMatch);
// 	    }
// 	}
// 	Advance(lgth);
//
// #endif
// #if defined(OMISSION) and ((OMISSION == END_TAG_MATCHING) or (OMISSION == NAME_LOOKUP))
// 	ScanTo(NameFollow);
// #endif
// //	for(int i=0; i<lgth; i++) {
// //		if (start_elem_name[i] != end_elem_name[i])
// //			WF_Error(wfErr_GIMatch);
// //	}
// //	if (start_elem_name[lgth] != '\0') WF_Error(wfErr_GIMatch);
//
// 	if (AtChar<ASCII,'>'>(cur())) {
// 		Advance(1);
// 		Parser_Interface<UTF_8>::EndTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
// 	}
//     else {
// 		ScanTo(NonWS);
// 		if (AtChar<ASCII,'>'>(cur())) {
// 			Advance(1);
// 			Parser_Interface<UTF_8>::EndTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
// 		}
// 		else Syntax_Error(NT_ETag);
//     }
// }


/* Parse a valid start or empty element tag. */
template <class B, WorkingCharacterSet W>
int ParsingEngine<B, W>::Parse_WF_StartTag (bool& is_emptyStartTag){
	int att_name_start;
	int att_val_start;
	int att_name_end, att_val_end;
	unsigned char quoteCh;
	Advance(1);

	#if (not defined(OMISSION)) or (OMISSION != NAME_LOOKUP)
	int nameID = Parse_Name();
	#endif
	#if (defined(OMISSION)) and (OMISSION == NAME_LOOKUP)
	ScanTo(NameFollow);
	int nameID = 0;
	#endif
	ElementName_action(GetCodeUnitPtr(text_or_markup_start+1), LengthFrom(text_or_markup_start+1));
	/* The following test optimizes the most common case of a
	start tag with no attributes.  */
	if (AtChar<B::Base,'>'>(cur())) {
		Advance(1);
		StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
	}
	else {
		ScanTo(NonWS);
		if (AtChar<B::Base,'>'>(cur())) {
			Advance(1);
			StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else if (at_EmptyElementDelim<B::Base>(cur())) {
			Advance(2);
			is_emptyStartTag = true;
			EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else do {
			/* Must be an attribute-value pair or error. */
			att_name_start = AbsPos();
			#if (not defined(OMISSION)) or (OMISSION != NAME_LOOKUP)
			int att_nameID = Parse_Name();
			#endif
			#if (defined(OMISSION)) and (OMISSION == NAME_LOOKUP)
			ScanTo(NameFollow);
			int att_nameID = 0;
			#endif
            att_name_end = AbsPos();
		#if (not defined(OMISSION)) or ((OMISSION != ATTRIBUTE_UNIQUENESS) and (OMISSION != NAME_LOOKUP))
			int attID = Parser_Interface<W>::model_info->getOrInsertGlobalAttName(att_nameID);
 			if (attID >= LastAttOccurrence.size()) LastAttOccurrence.push_back(0);
 			else {
 				if (LastAttOccurrence[attID] > text_or_markup_start) {
					WF_Error(wfErr_uniqattspec); /* Duplicate attribute. */
					break;
 				}
 			}
 		 	LastAttOccurrence[attID] = att_name_start;
 		 #endif
			/* The following optimized tests handle the frequently occurring
			case that there are no blanks on either side of the equals sign.
			In many cases, the very first test handles 100% of actual
			attribute-value pairs encountered. */
			if (at_EqualsQuote<B::Base>(cur())) Advance(1);
			else {
				ScanTo(NonWS);
				if (!AtChar<B::Base,'='>(cur())) {
					Syntax_Error(NT_STag);
					break;
				}
				Advance(1);
				ScanTo(NonWS);
				if (!AtQuote<B::Base>(cur())) {
					Syntax_Error(NT_STag);
					break;
				}
			}
			att_val_start = AbsPos()+1;
			Parse_AttValue();
			att_val_end = AbsPos()-1;
			if (at_xmlns<B::Base>(cur()+att_name_start-AbsPos())) {
				Namespace_action(GetCodeUnitPtr(att_name_start), att_name_end - att_name_start,
						 GetCodeUnitPtr(att_val_start), att_val_end - att_val_start);
			}
			else {
				AttributeValue_action(GetCodeUnitPtr(att_name_start), att_name_end - att_name_start,
						 GetCodeUnitPtr(att_val_start), att_val_end - att_val_start);
			}
			/* Now check for end or repeat. Avoid whitespace scan if possible.*/
			if (AtChar<B::Base,'>'>(cur())) {
				Advance(1);
				StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (at_EmptyElementDelim<B::Base>(cur())) {
				Advance(2);
				is_emptyStartTag = true;
				EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			ScanTo(NonWS);
			if (AtChar<B::Base,'>'>(cur())) {
				Advance(1);
				StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (at_EmptyElementDelim<B::Base>(cur())) {
				Advance(2);
				is_emptyStartTag = true;
				EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (AbsPos() == att_val_end + 1) {
				/* No WS following att value */
				Syntax_Error(NT_STag);
				break;
			}
		} while (1);
	}
	return nameID;
}



template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_WF_Element() {
	bool is_emptyStartTag = false;
	int nameID = Parse_WF_StartTag(is_emptyStartTag);
#ifdef DEBUG
	printf("Parse_Element: nameID = %d, is_emptyStartTag=%i\n",nameID, is_emptyStartTag);
#endif
	if (!is_emptyStartTag) {
		Parse_WF_Content();
		Parse_WF_EndTag(nameID);
	}
}


template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_WF_Content() {
	do {
		text_or_markup_start = AbsPos();
		ScanTextTo(MarkupStart); /* '<', '&', or ']' for ']]>' test */
		if (at_ElementTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_WF_Element();
		}
		else if (at_EndTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			return;
		}
		else if (at_Comment_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_Comment();
		}
		else if (at_CharRef_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CharRef();
		}
		else if (AtChar<B::Base,'&'>(cur())) {
			text_if_nonnull_action(true);
			Parse_EntityRef();
		}
		else if (at_CDATA_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CDATA();
		}
		else if (at_PI_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_PI();
		}
		else if (at_CDATA_End<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Advance(3);
			Syntax_Error(NT_CharData);
		}
		else if (at_EOF()) {
			text_if_nonnull_action(false);
			return;
		}
		else if (AtChar<B::Base,'<'>(cur())) {
			Syntax_Error(NT_markupdecl);
		}
		else {
			Advance(1);
			continue;
		}
	} while (1);
}


#ifndef MARKUP_PASS_CONTROL
#ifndef MARKUP_SORTING
template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::ParseContent() {
	Parser_Interface<W>::DocumentStart_action();
	bool is_emptyStartTag = false;
	do {
		text_or_markup_start = AbsPos();
		ScanTextTo(MarkupStart); /* '<', '&', or ']' for ']]>' test */
/*		if (AtChar<B::Base,'<'>(cur())) {
			text_if_nonnull_action();
			Parse_Markup<B, W>();
		}*/
		if (at_ElementTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_StartTag();
		}
		else if (at_EndTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_EndTag();
		}
		else if (at_Comment_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_Comment();
		}
		else if (at_CharRef_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CharRef();
		}
		else if (AtChar<B::Base,'&'>(cur())) {
			text_if_nonnull_action(true);
			Parse_EntityRef();
		}
		else if (at_CDATA_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CDATA();
		}
		else if (at_PI_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_PI();
		}
		else if (at_CDATA_End<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Advance(3);
			Syntax_Error(NT_CharData);
		}
		else if (at_EOF()) {
			text_if_nonnull_action(false);
			break;
		}
		else if (AtChar<B::Base,'<'>(cur())) {
			Syntax_Error(NT_markupdecl);
		}
		else {
			Advance(1);
			continue;
		}
	} while (1);
	Parser_Interface<W>::DocumentEnd_action();
}
#endif
#endif

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_DocType (){

	int old_abspos, start_pos;
	ScanTo(NonWS);
	start_pos = AbsPos();

	if (at_DOCTYPE_start<B::Base>(cur()))
    	Advance(9);
	else{
//		printf("No Document definition!\n");
		return;
	}
	requireWS();
	int nameID = Parse_Name();

	old_abspos = AbsPos();
    ScanTo(NonWS);
    if(at_SYSTEM<B::Base>(cur())||at_PUBLIC<B::Base>(cur())){
    	Parser_Interface<W>::model_info->has_external_DTD = true;
    	if(old_abspos == AbsPos())
    		Syntax_Error(NT_doctypedecl);
    	Parse_ExternalID(Parser_Interface<W>::model_info->external_DTD_systemLiteral, Parser_Interface<W>::model_info->external_DTD_pubidLiteral);
    	Parser_Interface<W> * entity_parser;
    	entity_parser = ParserFactory(Parser_Interface<W>::model_info->external_DTD_systemLiteral, Parser_Interface<W>::model_info);
		entity_parser->Parse_ExtSubsetDecl();
		entity_parser->~Parser_Interface<W>();
    }
    else Parser_Interface<W>::model_info->has_external_DTD = false;
    ScanTo(NonWS);

	if (AtChar<B::Base,'['>(cur())){
		Advance(1);
		Parse_IntSubset();
		if (AtChar<B::Base,']'>(cur()))
			Advance(1);
		else
    		Syntax_Error(NT_doctypedecl);
		ScanTo(NonWS);
	}

	if (AtChar<B::Base,'>'>(cur())){
		Advance(1);

		CRE_Seq * rslt = new CRE_Seq();
		rslt->subCMs.push_back(new CRE_Name(nameID));
		CM_RegExp * cre = new CM_RegExp();
		cre->content_re = rslt;

		int id_count = cre->content_re->Set_IDs(0);
		cre->content_re->Set_First_Map();
		symbol_set_t * transition_map = new symbol_set_t[id_count+1];
		cre->content_re->follow_map[0] = id_count+1;

		cre->content_re->Set_Follow_Map(transition_map);
		transition_map[0] = cre->content_re->first_map;
		if (cre->content_re->matches_empty)
			transition_map[0][0]=id_count+1;

		cre -> transition_map = transition_map;

		Parser_Interface<W>::model_info->rootModel = cre;

		/* Check for notations that were used, but not defined by the end of the DTD. */
		#if (VALIDATION_MODE == ON)
		hash_map<int, int >::iterator j;
		for (j=Parser_Interface<W>::model_info->GlobalNotationTable.begin(); j!=Parser_Interface<W>::model_info->GlobalNotationTable.end(); j++) {
			if (j->second == -1)
				Validity_Error(vErr_notatn);
		}
		#endif
	}
	else
		Syntax_Error(NT_doctypedecl);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_ExternalID (char *& systemLiteral, char *& pubidLiteral){
	int quot_start, lgth;
	if(at_SYSTEM<B::Base>(cur())){
		Advance(6);
		pubidLiteral = NULL;
		requireWS();
		if (!AtQuote<B::Base>(cur())) Syntax_Error(NT_ExternalID);
		quot_start = AbsPos()+1;
		Parse_SystemLiteral (); /*  SystemLiteral */
		lgth = AbsPos() - quot_start - 1;
		systemLiteral = copy_string(GetCodeUnitPtr(quot_start),lgth);
	}
	else if (at_PUBLIC<B::Base>(cur())){
		Advance(6);
		requireWS();
		if (!AtQuote<B::Base>(cur())) Syntax_Error(NT_ExternalID);
		quot_start = AbsPos()+1;
		Parse_PubidLiteral ();/*  PubidLiteral */
		lgth = AbsPos() - quot_start - 1;
		pubidLiteral = copy_string(GetCodeUnitPtr(quot_start),lgth);
		systemLiteral = NULL;
		if (AtChar<B::Base, '>'>(cur())) return;
		requireWS();
		if (AtQuote<B::Base>(cur())) {
			quot_start = AbsPos()+1;
			Parse_SystemLiteral ();/*  SystemLiteral */
			lgth = AbsPos() - quot_start - 1;
			systemLiteral = copy_string(GetCodeUnitPtr(quot_start),lgth);
		}
	}
	else
		Syntax_Error(NT_ExternalID);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_SystemLiteral (){
	unsigned char quoteCh;
	if(AtQuote<B::Base>(cur())){
		quoteCh = cur()[0];
		Advance(1);
	}
	ScanTo(Quote);
	while (cur()[0] != quoteCh){
		if(at_EOF())
			Syntax_Error(NT_SystemLiteral);
		Advance(1);
		ScanTo(Quote);
	}
	Advance(1);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_PubidLiteral (){
	unsigned char quoteCh;
	quoteCh = cur()[0];
	Advance(1);
	while (at_PubidChar<B::Base>(cur()) && (cur()[0] != quoteCh)) {
		Advance(1);
	}
	if (cur()[0] != quoteCh){
		Syntax_Error(NT_PubidLiteral);
	}
	Advance(1);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_IntSubset (){

	while(1){
		ScanTo(NonWS);
		text_or_markup_start = AbsPos();
		if (AtChar<B::Base,'%'>(cur()))
			Parse_PEReference();
		else if (at_PI_Start<B::Base>(cur())) {
			Parse_PI();
		}
		else if (at_Comment_Start<B::Base>(cur())) {
			Parse_Comment();
		}
		else if (AtChar<B::Base,'<'>(cur())){
			Advance(1);
			if(AtChar<B::Base,'!'>(cur())){
				Advance(1);
				if (at_ELEMENT<B::Base>(cur()))
					Parse_Elementdecl();
				else if (at_ATTLIST<B::Base>(cur()))
					Parse_AttlistDecl();
				else if (at_ENTITY<B::Base>(cur()))
					Parse_Entitydecl();
				else if (at_NOTATION<B::Base>(cur()))
					Parse_Notationdecl();
				else {
					Syntax_Error(NT_markupdecl);
				}
			}
			else
				Syntax_Error(NT_markupdecl);
		}
		else if (AtChar<B::Base,']'>(cur())){
			break;
		}
		else
			Syntax_Error(NT_intSubset);
	}
}


template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_PEReference (){

	Advance(1); /* Skip "%". */
	fprintf(stderr,"Parameter Reference has not been completed yet.\n");
	exit(-1);
	int nameID = Parse_Name();
	if (AtChar<B::Base,';'>(cur())) {
		Advance(1);
		PEReference_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		PEntity_info * this_info;
		Parser_Interface<W> * entity_parser;
		int entityID = Parser_Interface<W>::model_info->GlobalPEntityTable[nameID];
		if (entityID == 0)
			WF_Error(wfErr_wf_entdeclared);
		else{
			this_info = Parser_Interface<W>::model_info->PEntityData[entityID-1];
			if (this_info->is_external){

//	     		if (Parser_Interface<W>::entity_Info->standalone != Standalone_no)
//	     			WF_Error(wfErr_NoExternalRefs);
//	     		else {
					entity_parser = ParserFactory(this_info->systemLiteral, Parser_Interface<W>::model_info);
					entity_parser->Parse_WF_Content();
					if(!entity_parser->at_EOF())
						Syntax_Error(NT_content);
					entity_parser->~Parser_Interface<W>();
//	     		}
			}
			else {
			}
		}
	}
	else
		Syntax_Error(NT_PEReference);
}


template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_Elementdecl (){

	Advance(7); /* Skip "<!ELEMENT". */

    requireWS();
	int nameID = Parse_Name();
	int elemID = Parser_Interface<W>::model_info->getOrInsertGlobalElement(nameID);

	requireWS();
	ContentModel * cm;
	/* Start parsing "contentspec"*/
	if (at_EMPTY<B::Base>(cur())) {
    	Advance(5);
    	cm = new CM_Empty();
    	Parser_Interface<W>::model_info->ContentModelData[nameID] = cm;
	}
    else if (at_ANY<B::Base>(cur())) {
    	Advance(3);
    	cm = new CM_Any();
    	Parser_Interface<W>::model_info->ContentModelData[nameID] = cm;
    }
    else {
    	if (AtChar<B::Base,'('>(cur()))
			Advance(1);
		ScanTo(NonWS);
		if (at_PCDATA<B::Base>(cur())){
			cm = Parse_RemainingMixed();
			Parser_Interface<W>::model_info->ContentModelData[nameID] = cm;
		}
		else{

			CM_RegExp * cre = new CM_RegExp;
			cre->content_re = Parse_RemainingChildren();

			int id_count = cre->content_re->Set_IDs(0);
			cre->content_re->Set_First_Map();
			symbol_set_t * transition_map = new symbol_set_t[id_count+1];
			cre->content_re->follow_map[0] = id_count+1;

			cre->content_re->Set_Follow_Map(transition_map);
			transition_map[0] = cre->content_re->first_map;

			if (cre->content_re->matches_empty)
				transition_map[0][0]=id_count+1;

			cre -> transition_map = transition_map;

			Parser_Interface<W>::model_info->ContentModelData[nameID] = cre;
			cm = cre;
		}
    }
    ScanTo(NonWS);

	if (AtChar<B::Base,'>'>(cur())) {
		Advance(1);
	}
	else
		Syntax_Error(NT_elementdecl);
}
template <class B, WorkingCharacterSet W>
ContentModel * ParsingEngine<B, W>::Parse_RemainingMixed (){
	CM_Mixed * r = new CM_Mixed();
	Advance(7);  /* Skip "#PCDATA". */

    if (AtChar<B::Base,')'>(cur())){
    	if (AtChar<B::Base,'*'>(cur())) {
    		Advance(2);
	   	}
		else {
			Advance(1);
		}
    }
    else{
    	ScanTo(NonWS);
    	int k = 0;
    	while (AtChar<B::Base,'|'>(cur())){
			Advance(1);
			ScanTo(NonWS);
			int nameID = Parse_Name();
			r->elements[nameID] = ++k;
			ScanTo(NonWS);
		}
		if (at_Para_star<B::Base>(cur())) Advance(2);
		else {
			Syntax_Error(NT_Mixed);
			exit(-1);
    	}
    }
    return r;
}


template <class B, WorkingCharacterSet W>
Content_RE * ParsingEngine<B, W>::Parse_RemainingChildren (){
	Content_RE * c1 = Parse_Cp();
	Content_RE * r = c1;
	ScanTo(NonWS);
	if(AtChar<B::Base,'|'>(cur())){
		CRE_Choice * rslt = new CRE_Choice;
		rslt->subCMs.push_back(c1);
		Advance(1);
		ScanTo(NonWS);
		rslt->subCMs.push_back(Parse_Cp());
		ScanTo(NonWS);
		while(!AtChar<B::Base,')'>(cur())){
			if(AtChar<B::Base,'|'>(cur()))
				Advance(1);
			else
				Syntax_Error(NT_children);
			ScanTo(NonWS);
			rslt->subCMs.push_back(Parse_Cp());
			ScanTo(NonWS);
		}
		Advance(1);
		rslt->Compile();
		r = rslt;
	}
	else if(AtChar<B::Base,','>(cur())){
		CRE_Seq * rslt = new CRE_Seq;
		rslt->subCMs.push_back(c1);
		Advance(1);
		ScanTo(NonWS);
		rslt->subCMs.push_back(Parse_Cp());
		ScanTo(NonWS);
		while(!AtChar<B::Base,')'>(cur())){
			if(AtChar<B::Base,','>(cur()))
				Advance(1);
			else
				Syntax_Error(NT_children);
			ScanTo(NonWS);
			rslt->subCMs.push_back(Parse_Cp());
			ScanTo(NonWS);
		}
		Advance(1);
		rslt->Compile();
		r = rslt;
	}
	else if(AtChar<B::Base,')'>(cur())){
		Advance(1);
	}
	else
		Syntax_Error(NT_children);

	if (AtChar<B::Base,'?'>(cur())) {
		Advance(1);
		r = new CRE_Opt(r);
	}
	else if (AtChar<B::Base,'*'>(cur())) {
		Advance(1);
		r = new CRE_Star(r);
	}
	else if (AtChar<B::Base,'+'>(cur())) {
		Advance(1);
		r = new CRE_Plus(r);
	}

	return r;
}

template <class B, WorkingCharacterSet W>
Content_RE * ParsingEngine<B, W>::Parse_Cp (){
	if (AtChar<B::Base,'('>(cur())){
		Advance(1);
		ScanTo(NonWS);
		Parse_RemainingChildren();
	}
	else{
		int nameID = Parse_Name();
		CRE_Name * r = new CRE_Name(nameID);

		if (AtChar<B::Base,'?'>(cur())) {
			Advance(1);
			return new CRE_Opt(r);
		}
		else if (AtChar<B::Base,'*'>(cur())) {
			Advance(1);
			return new CRE_Star(r);
		}
		else if (AtChar<B::Base,'+'>(cur())) {
			Advance(1);
			return new CRE_Plus(r);
		}
		else return r;
	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_AttlistDecl (){

	int old_abspos;

	int name_start;
	int lgth;

	int elemID;
	int attID;

	Advance(7); /* Skip "ATTLIST. */
	requireWS();

	int nameID = Parse_Name();
	elemID = Parser_Interface<W>::model_info->getOrInsertGlobalElement(nameID);

	old_abspos = AbsPos();
	ScanTo(NonWS);
	while(!AtChar<B::Base,'>'>(cur())) {
		if(old_abspos == AbsPos())
    		Syntax_Error(NT_AttlistDecl);

		int att_nameID = Parse_Name();

		attID = Parser_Interface<W>::model_info->getOrInsertGlobalAttName(att_nameID);
 		if (attID >= LastAttOccurrence.size()) LastAttOccurrence.push_back(0);
    	ATT_info * this_info = new ATT_info;
    	this_info->globalATT_id = attID;
    	requireWS();
    	if (at_CDATA<B::Base>(cur())){
    		Advance(5);
    		this_info->attType = CDATA_att;
    	}
    	else if(at_ID<B::Base>(cur())){
    		Advance(2);
    		this_info->attType = ID_att;
    	}
    	/* Make sure to check IDREFS before IDREF*/
    	else if(at_IDREFS<B::Base>(cur())){
    		Advance(6);
    		this_info->attType = IDREFS_att;
    	}
    	else if(at_IDREF<B::Base>(cur())){
    		Advance(5);
    		this_info->attType = IDREF_att;
    	}
    	else if(at_ENTITY<B::Base>(cur())){
    		Advance(6);
    		this_info->attType = ENTITY_att;
    	}
    	else if(at_ENTITIES<B::Base>(cur())){
    		Advance(8);
    		this_info->attType = ENTITIES_att;
    	}
    	/* Make sure to check NMTOKENS before NMTOKEN*/
    	else if(at_NMTOKENS<B::Base>(cur())){
    		Advance(8);
    		this_info->attType = NMTOKENS_att;
    	}
    	else if(at_NMTOKEN<B::Base>(cur())){
    		Advance(7);
    		this_info->attType = NMTOKEN_att;
    	}
    	else if(at_NOTATION<B::Base>(cur())){ /* NotationType = 'NOTATION' S Enumeration
    									 when Nmtoken = Name */
    		Advance(8);
			requireWS();
    		Parse_Notation(this_info);
    		this_info->attType = NOTATION_att;
    	}
    	else if(AtChar<B::Base,'('>(cur())){
    		Parse_Enumeration(this_info);
    		this_info->attType = enumeration_att;
    	}
    	else
    		Syntax_Error(NT_AttlistDecl);
    	requireWS();
    	Parse_DefaultDecl(this_info);

		ScanTo(NonWS);
		Parser_Interface<W>::model_info->ElementAttributeData[elemID].push_back(this_info);
	}

	Advance(1);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_Notation (ATT_info * this_info){

	if(AtChar<B::Base,'('>(cur()))
		Advance(1);
	else
		Syntax_Error(NT_NotationType);
	ScanTo(NonWS);

    int notn_nameID = Parse_Name();

	/*Notation name is not in the global table!*/
	if(Parser_Interface<W>::model_info->GlobalNotationTable[notn_nameID]==0)
		Parser_Interface<W>::model_info->GlobalNotationTable[notn_nameID] = -1;

	ScanTo(NonWS);
	while(AtChar<B::Base,'|'>(cur())){
		Advance(1);
		ScanTo(NonWS);
		notn_nameID = Parse_Name();

		if(Parser_Interface<W>::model_info->GlobalNotationTable[notn_nameID]==0)
//			Validity_Error(vErr_notatn);
			Parser_Interface<W>::model_info->GlobalNotationTable[notn_nameID] = -1;

		ScanTo(NonWS);
	}
	if (AtChar<B::Base,')'>(cur()))
		Advance(1);
	else
		Syntax_Error(NT_NotationType);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_Enumeration (ATT_info * this_info){

	int enumCount=0;
	if(AtChar<B::Base,'('>(cur()))
		Advance(1);
	else
		Syntax_Error(NT_Enumeration);
	ScanTo(NonWS);

	int nmtokenID = Parse_Nmtoken();

	this_info->enumValues[nmtokenID]=++(enumCount);

	ScanTo(NonWS);
	while(AtChar<B::Base,'|'>(cur())){
		Advance(1);
		ScanTo(NonWS);
		int nmtokenID = Parse_Nmtoken();

		int enumID = this_info->enumValues[nmtokenID];
		if(enumID==0){
			this_info->enumValues[nmtokenID]=++(enumCount);
			enumID = enumCount;
		}
		else if(!StrictWellFormedness){
			Validity_Error(vErr_NoDuplicateTokens);
		}
		ScanTo(NonWS);
	}
	if (AtChar<B::Base,')'>(cur()))
		Advance(1);
	else
		Syntax_Error(NT_Enumeration);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_DefaultDecl (ATT_info * this_info){
	if(at_REQUIRED<B::Base>(cur())){
		Advance(9);
		this_info->defaultKind = REQUIRED_att;
	}
	else if(at_IMPLIED<B::Base>(cur())){
		Advance(8);
		this_info->defaultKind = IMPLIED_att;
	}
	else {
		if(at_FIXED<B::Base>(cur())){
			Advance(6);
			requireWS();
			this_info->defaultKind = FIXED_att;
		}
		else this_info->defaultKind = DEFAULT_att;
		if(AtQuote<B::Base>(cur())){
			int quot_start = AbsPos()+1;
			Parse_AttValue();
			/* need to normalize */
			this_info->defaultValueLgth = AbsPos() - quot_start - 1;

			this_info->defaultValue = new unsigned char[this_info->defaultValueLgth+1];
			memcpy(this_info->defaultValue, GetCodeUnitPtr(quot_start),this_info->defaultValueLgth);
			this_info->defaultValue[this_info->defaultValueLgth] = '\0';
			}
		else
			Syntax_Error(NT_DefaultDecl);
	}
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_Entitydecl (){

	int name_start;
	int quot_start;
	int lgth;
	int old_abspos;
	char * s;

	Advance(6); /* Skip "ENTITY. */
	requireWS();

	if (AtChar<B::Base,'%'>(cur())){
		Advance(1);
		requireWS();

		int nameID = Parse_Name();
		PEntity_info * this_info = new PEntity_info;
		int entityID = Parser_Interface<W>::model_info->GlobalPEntityTable[nameID];
		if(entityID==0){
			Parser_Interface<W>::model_info->GlobalPEntityTable[nameID]=++(Parser_Interface<W>::model_info->globalPEntityCount);
			entityID = Parser_Interface<W>::model_info->globalPEntityCount;
			this_info->globalPEntity_id = entityID;
		}
		else
			printf("Warning: Entity definition already exist!\n");

		requireWS();
		if(AtQuote<B::Base>(cur())){
    		Parse_PEntityValue(this_info);
    		this_info->is_external = false;
    	}
    	else {
     		Parse_ExternalID(this_info->systemLiteral, this_info->pubidLiteral);
     		this_info->is_external = true;
     		if (this_info->systemLiteral == NULL) Syntax_Error(NT_EntityDecl);
    	}
     	Parser_Interface<W>::model_info->PEntityData.push_back(this_info);
	}
	else{
		int nameID = Parse_Name();

		GEntity_info * this_info = new GEntity_info();
		int entityID = Parser_Interface<W>::model_info->GlobalGEntityTable[nameID];
		if(entityID==0){
			Parser_Interface<W>::model_info->GlobalGEntityTable[nameID]=++(Parser_Interface<W>::model_info->globalGEntityCount);
			entityID = Parser_Interface<W>::model_info->globalGEntityCount;
			this_info->globalGEntity_id = entityID;
		}
		else
			printf("Warning: Entity definition already exists!\n");

		requireWS();

		if(AtQuote<B::Base>(cur())){
    		Parse_GEntityValue(this_info);
    		this_info->is_external = false;
    	}
    	else {
     		Parse_ExternalID(this_info->systemLiteral, this_info->pubidLiteral);
     		this_info->is_external = true;
    		if (this_info->systemLiteral == NULL) Syntax_Error(NT_EntityDecl);
			old_abspos = AbsPos();
			ScanTo(NonWS);
    	   	if(at_NDATA<B::Base>(cur())){
    	   		if(old_abspos == AbsPos())
    				Syntax_Error(NT_EntityDecl);
    			else
    				Advance(5);
    			requireWS();
    			name_start = AbsPos();
    			int nameID = Parse_Name();
    			lgth = AbsPos() - name_start;
				this_info->NDataName = copy_string(GetCodeUnitPtr(name_start),lgth);
    	   	}
		}
    	Parser_Interface<W>::model_info->GEntityData.push_back(this_info);
	}
	ScanTo(NonWS);
	if (AtChar<B::Base,'>'>(cur())){
		Advance(1);
	}
	else
		Syntax_Error(NT_EntityDecl);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_Notationdecl (){

	int old_abspos;
	Advance(8); /* Skip "NOTATION. */
	requireWS();

	int nameID = Parse_Name();

	int notationID = Parser_Interface<W>::model_info->GlobalNotationTable[nameID];
	/* notationID == -1: used but not yet defined; == 0: new, > 0 prev. defined */
	if(notationID <= 0){
		Parser_Interface<W>::model_info->GlobalNotationTable[nameID]=++(Parser_Interface<W>::model_info->globalNotationCount);
		notationID = Parser_Interface<W>::model_info->globalNotationCount;
	}
	else /*Duplicate notation name!*/
		Validity_Error(vErr_NoDuplicateTokens);
	Notation_info * this_info = new Notation_info;
	ScanTo(NonWS);
    Parse_ExternalID(this_info->systemLiteral, this_info->pubidLiteral);
	ScanTo(NonWS);
	if (AtChar<B::Base,'>'>(cur())) {
		Advance(1);
	}
	else
		Syntax_Error(NT_NotationDecl);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::requireWS(){

    int old_abspos = AbsPos();
    ScanTo(NonWS);
    if(old_abspos == AbsPos())
    	Syntax_Error(NT_S);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_AttValue(){

	int	quoteCh = cur()[0];
	Advance(1); /* Skip " or ' */

	ScanTo(Quote);
	while (cur()[0] != quoteCh){
		if (at_CharRef_Start<B::Base>(cur())){
			Parse_CharRef();
			ScanTo(Quote);
		}
		else if (AtChar<B::Base,'&'>(cur())){
			Parse_EntityRef();
			ScanTo(Quote);
		}
		else if (AtQuote<B::Base>(cur())) {
			Advance(1);
			ScanTo(Quote);
		}
		else /* if (AtChar<B::Base,'<'>(cur())) */
			WF_Error(wfErr_CleanAttrVals);
	}
	Advance(1);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_GEntityValue(GEntity_info * this_info){

	int	quoteCh = cur()[0];
	Advance(1); /* Skip " or ' */
	this_info->is_simple = true;
	int quot_start = AbsPos();
	char * replText;
	ScanTo(Quote);
	replText = copy_string(GetCodeUnitPtr(quot_start),AbsPos()-quot_start);
	while (cur()[0] != quoteCh){
		if (at_CharRef_Start<B::Base>(cur())){
			strcat (replText,Replace_CharRef());
			quot_start = AbsPos();
			ScanTo(Quote);
		}
		else if (AtQuote<B::Base>(cur())) {
			quot_start = AbsPos();
			Advance(1);
			ScanTo(Quote);
		}
		else if (at_EOF()) {
			Syntax_Error(NT_EntityValue);
		}
		else { /* '<' or '&' found */
			quot_start = AbsPos();
			Advance(1);
			ScanTo(Quote);
			this_info->is_simple = false;
		}
		replText = cat_string (replText,(char *)GetCodeUnitPtr(quot_start), strlen(replText), AbsPos()-quot_start);
	}
	this_info->ReplacementText = replText;
	Advance(1);
}

template <class B, WorkingCharacterSet W>
char * ParsingEngine<B, W>::Replace_EntityRef(bool& is_simple){
	Advance(1);
	int nameID = Parse_Name();
	if (AtChar<B::Base,';'>(cur()))
		Advance(1);
	else
		Syntax_Error(NT_EntityValue);
	int entityID = Parser_Interface<W>::model_info->GlobalGEntityTable[nameID];
	if (entityID == 0)
		WF_Error(wfErr_wf_entdeclared);
	else{
		if (Parser_Interface<W>::model_info->GEntityData[entityID-1]->is_simple == false)
			is_simple = false;
		return Parser_Interface<W>::model_info->GEntityData[entityID-1]->ReplacementText;
	}

}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_PEntityValue(PEntity_info * this_info){
	fprintf(stderr,"parsing of parameter entity value has not been completed yet.\n");
	exit(-1);
}

template <class B, WorkingCharacterSet W>
char * ParsingEngine<B, W>::Replace_CharRef(){
	Advance(2);
	fprintf(stderr,"Replacement of Character Reference has not been completed yet.\n");
	exit(-1);
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_Prolog(){
	ScanTo(NonWS);
	int old_pos = AbsPos();
	while (!at_DOCTYPE_start<B::Base>(cur())) {
		text_or_markup_start = AbsPos();
		if (at_Comment_Start<B::Base>(cur()))
			Parse_Comment();
		else if (at_PI_Start<B::Base>(cur()))
				Parse_PI();
		else{
			Prolog_action(GetCodeUnitPtr(old_pos), LengthFrom(old_pos));
			return;
		}
		ScanTo(NonWS);
	}
	Parse_DocType();
	ScanTo(NonWS);
	while(at_Comment_Start<B::Base>(cur()) || at_PI_Start<B::Base>(cur()) ){
		text_or_markup_start = AbsPos();
		if (at_Comment_Start<B::Base>(cur()))
			Parse_Comment();
		else
			Parse_PI();
		ScanTo(NonWS);
	}
	Prolog_action(GetCodeUnitPtr(old_pos), LengthFrom(old_pos));
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_ExtSubsetDecl() {
	ScanTo(NonWS);
	int start_pos=AbsPos();
	while(!at_EOF()){
		if(at_condSect_start<B::Base>(cur())){
			Advance(3);
			ScanTo(NonWS);
			if (at_INCLUDE<B::Base>(cur())){
				Advance(7);
				ScanTo(NonWS);
				if(AtChar<B::Base,'['>(cur())){
					Advance(1);
					Parse_ExtSubsetDecl();
					if(at_CDATA_End<B::Base>(cur()))
						Advance(3);
					else Syntax_Error(NT_includeSect);
				}
				else Syntax_Error(NT_includeSect);
			}
			else if (at_IGNORE<B::Base>(cur())){
				Advance(6);
				ScanTo(NonWS);
				if(AtChar<B::Base,'['>(cur())){
					int section_depth=1;
					Advance(1);
					while(!at_EOF()){
						ScanTextTo(MarkupStart);
						if(at_condSect_start<B::Base>(cur())){
							Advance(3);
							section_depth++;
						}
						else if(at_CDATA_End<B::Base>(cur())){
							Advance(3);
							section_depth--;
						}
						else
							Advance(1);
						if(section_depth==0) return;
					}
					Syntax_Error(NT_ignoreSectContents);
				}
				else Syntax_Error(NT_ignoreSect);
			}
			else Syntax_Error(NT_conditionalSect);
		}
		else if (AtChar<B::Base,'%'>(cur()))
			Parse_PEReference();
		else if (at_PI_Start<B::Base>(cur())) {
			Parse_PI();
		}
		else if (at_Comment_Start<B::Base>(cur())) {
			Parse_Comment();
		}
		else if (AtChar<B::Base,'<'>(cur())){
			Advance(1);

			if(AtChar<B::Base,'!'>(cur())){
				Advance(1);
				if(at_ELEMENT<B::Base>(cur()))
					Parse_Elementdecl();
				else if(at_ATTLIST<B::Base>(cur()))
					Parse_AttlistDecl();
				else if(at_ENTITY<B::Base>(cur()))
					Parse_Entitydecl();
				else if(at_NOTATION<B::Base>(cur()))
					Parse_Notationdecl();
				else{
					Syntax_Error(NT_markupdecl);
				}
			}
			else
				Syntax_Error(NT_markupdecl);
		}
		else
			Syntax_Error(NT_extSubsetDecl);
		ScanTo(NonWS);
	}
	ExtSubsetDecl_action(GetCodeUnitPtr(start_pos), LengthFrom(start_pos));
}

/* Parse a valid start or empty element tag. */
template <class B, WorkingCharacterSet W>
inline int ParsingEngine<B, W>::Parse_ValidStartTag (bool& is_emptyStartTag){
	int att_name_start;
	int att_val_start;
	int att_name_end, att_val_end;
	unsigned char quoteCh;
	Advance(1);

	int nameID = Parse_Name();
	int elemID = Parser_Interface<W>::model_info->GlobalElementTable[nameID];
	if(elemID==0)
			Validity_Error(vErr_elementvalid);

	ElementName_action(GetCodeUnitPtr(text_or_markup_start+1), LengthFrom(text_or_markup_start+1));
	/* The following test optimizes the most common case of a
	start tag with no attributes.  */
	if (AtChar<B::Base,'>'>(cur())) {
		Advance(1);
		StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
	}
	else {
		ScanTo(NonWS);
		if (AtChar<B::Base,'>'>(cur())) {
			Advance(1);
			StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else if (at_EmptyElementDelim<B::Base>(cur())) {
			Advance(2);
			is_emptyStartTag = true;
			EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
		}
		else do {
			/* Must be an attribute-value pair or error. */
			att_name_start = AbsPos();
			int att_nameID = Parse_Name();
			#if (not defined(OMISSION)) or (OMISSION != ATTRIBUTE_UNIQUENESS)
			int attID = Parser_Interface<W>::model_info->getOrInsertGlobalAttName(att_nameID);
 			if (attID >= LastAttOccurrence.size()) LastAttOccurrence.push_back(0);
 			else {
 				if (LastAttOccurrence[attID] > text_or_markup_start) {
					WF_Error(wfErr_uniqattspec); /* Duplicate attribute. */
					break;
 				}
 			}
 		 	LastAttOccurrence[attID] = att_name_start;
 		 	#endif
			/* The following optimized tests handle the frequently occurring
			case that there are no blanks on either side of the equals sign.
			In many cases, the very first test handles 100% of actual
			attribute-value pairs encountered. */
			if (at_EqualsQuote<B::Base>(cur())) Advance(1);
			else {
				ScanTo(NonWS);
				if (!AtChar<B::Base,'='>(cur())) {
					Syntax_Error(NT_STag);
					break;
				}
				Advance(1);
				ScanTo(NonWS);
				if (!AtQuote<B::Base>(cur())) {
					Syntax_Error(NT_STag);
					break;
				}
			}
			att_val_start = AbsPos()+1;
			Parse_AttValue();
			att_val_end = AbsPos()-1;
			if (at_xmlns<B::Base>(cur()+att_name_start-AbsPos())) {
				Namespace_action(GetCodeUnitPtr(att_name_start), att_name_end - att_name_start,
						 GetCodeUnitPtr(att_val_start), att_val_end - att_val_start);
			}
			else {
				AttributeValue_action(GetCodeUnitPtr(att_name_start), att_name_end - att_name_start,
						 GetCodeUnitPtr(att_val_start), att_val_end - att_val_start);
			}
			/* Now check for end or repeat. Avoid whitespace scan if possible.*/
			if (AtChar<B::Base,'>'>(cur())) {
				Advance(1);
				StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (at_EmptyElementDelim<B::Base>(cur())) {
				Advance(2);
				is_emptyStartTag = true;
				EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			ScanTo(NonWS);
			if (AtChar<B::Base,'>'>(cur())) {
				Advance(1);
				StartTag_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (at_EmptyElementDelim<B::Base>(cur())) {
				Advance(2);
				is_emptyStartTag = true;
				EmptyElement_action(GetCodeUnitPtr(text_or_markup_start), LengthFrom(text_or_markup_start));
				break;
			}
			else if (AbsPos() == att_val_end + 1) {
				/* No WS following att value */
				Syntax_Error(NT_STag);
				break;
			}
		} while (1);
	}
	return nameID;
}

template <class B, WorkingCharacterSet W>
int ParsingEngine<B, W>::Parse_ValidElement() {
	bool is_emptyStartTag = false;
	int nameID = Parse_ValidStartTag(is_emptyStartTag);
#ifdef DEBUG
	printf("Parse_ValidElement: nameID = %d, name = %s, is_emptyStartTag=%i\n",nameID, Parser_Interface<W>::model_info->symbol_table->Get_UTF8_name(nameID), is_emptyStartTag);
#endif
	ContentModel * cm = Parser_Interface<W>::model_info->ContentModelData[nameID];
	switch (cm->cm_type) {
		case cm_Empty:
			if (!is_emptyStartTag) {
				if (at_EndTag_Start<B::Base>(cur())) {
					Parse_WF_EndTag(nameID);
				}
				else {
					Validity_Error(vErr_elementvalid);
				}
			}
			break;
		case cm_Any:
			if (!is_emptyStartTag) {
				Parse_AnyContent();
				Parse_WF_EndTag(nameID);
			}
			break;
		case cm_Mixed:
			if (!is_emptyStartTag) {
				Parse_MixedContent(((CM_Mixed *) cm)->elements);
				Parse_WF_EndTag(nameID);
			}
			break;
		case cm_RegExp:
			CM_RegExp * cre = (CM_RegExp *) cm;
			int content_state = 0;
			if (!is_emptyStartTag) {
				Parse_ValidContent(cre, content_state);
				#ifdef DEBUG
				printf("Final content_state = %i, nameID = %i\n", content_state, nameID);
				#endif
				Parse_WF_EndTag(nameID);
			}
			if (cre->transition_map[content_state][0]==0) {
				Validity_Error(vErr_elementvalid);
			}
	}
	return nameID;
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_ValidContent(CM_RegExp * cre, int & cur_state) {
	do {
		ScanTo(NonWS);
		/* If non-null report WS  WS_action()? */
		text_or_markup_start = AbsPos();
		if (at_EndTag_Start<B::Base>(cur())) {
			break;
		}
		else if (at_ElementTag_Start<B::Base>(cur())) {
			int nameID = Parse_ValidElement();
#ifdef DEBUG
			printf("Content model state transition %i", cur_state);
#endif
			cur_state = cre->transition_map[cur_state][nameID];
#ifdef DEBUG
			printf("-> %i\n", cur_state);
#endif
		}
		else if (at_Comment_Start<B::Base>(cur())) {
			Parse_Comment();
		}
		else if (at_PI_Start<B::Base>(cur())) {
			Parse_PI();
		}
		else if (AtChar<B::Base,'&'>(cur())) {
			Parse_ValidEntityRef(cre, cur_state);
#ifdef DEBUG
			printf("EntityRef complete, cur_state = %i\n", cur_state);
#endif

		}
		else if (at_EOF()) {
			break;
		}
		else if (AtChar<B::Base,'<'>(cur())) {
			Syntax_Error(NT_markupdecl);
		}
		else {
			Validity_Error(vErr_elementvalid);
		}
	} while(1);
}


template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_AnyContent() {
	do {
		text_or_markup_start = AbsPos();
		ScanTextTo(MarkupStart); /* '<', '&', or ']' for ']]>' test */
		if (at_ElementTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			int nameID = Parse_ValidElement();
		}
		else if (at_EndTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			return;
		}
		else if (at_Comment_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_Comment();
		}
		else if (at_CharRef_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CharRef();
		}
		else if (AtChar<B::Base,'&'>(cur())) {
			text_if_nonnull_action(true);
			Parse_EntityRef_inAnyContent();
		}
		else if (at_CDATA_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CDATA();
		}
		else if (at_PI_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_PI();
		}
		else if (at_CDATA_End<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Advance(3);
			Syntax_Error(NT_CharData);
		}
		else if (at_EOF()) {
			text_if_nonnull_action(false);
			return;
		}
		else if (AtChar<B::Base,'<'>(cur())) {
			Syntax_Error(NT_markupdecl);
		}
		else {
			Advance(1);
			continue;
		}
	} while (1);
}
template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_MixedContent(symbol_set_t elems) {
	do {
		text_or_markup_start = AbsPos();
		ScanTextTo(MarkupStart); /* '<', '&', or ']' for ']]>' test */
/*		if (AtChar<B::Base,'<'>(cur())) {
			text_if_nonnull_action();
			Parse_Markup<B, W>();
		}*/
		if (at_ElementTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			int nameID = Parse_ValidElement();
			if (elems[nameID] == 0) {
				Validity_Error(vErr_elementvalid);
			}
		}
		else if (at_EndTag_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			return;
		}
		else if (at_Comment_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_Comment();
		}
		else if (at_CharRef_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CharRef();
		}
		else if (AtChar<B::Base,'&'>(cur())) {
			text_if_nonnull_action(true);
			Parse_EntityRef_inMixed(elems);
		}
		else if (at_CDATA_Start<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Parse_CDATA();
		}
		else if (at_PI_Start<B::Base>(cur())) {
			text_if_nonnull_action(false);
			Parse_PI();
		}
		else if (at_CDATA_End<B::Base>(cur())) {
			text_if_nonnull_action(true);
			Advance(3);
			Syntax_Error(NT_CharData);
		}
		else if (at_EOF()) {
			text_if_nonnull_action(false);
			return;
		}
		else if (AtChar<B::Base,'<'>(cur())) {
			Syntax_Error(NT_markupdecl);
		}
		else {
			Advance(1);
			continue;
		}
	} while (1);
}


template <class B, WorkingCharacterSet W>
int ParsingEngine<B, W>::Parse_Name() {
	int name_pos = AbsPos();
	ScanTo(NameFollow);
	int lgth = AbsPos()-name_pos;
	int nameID = Parser_Interface<W>::model_info->symbol_table->ASCII_Lookup_or_Insert_Name(&((char *) x8data)[buffer_rel_pos-lgth], lgth);
	if (nameID != 0) return nameID;
	else {
		int u8_lgth = byteplex->UTF8_Length(name_pos, lgth);
		char * u8_ptr = Parser_Interface<W>::model_info->symbol_table->ReserveSymbolSpace(u8_lgth);
		byteplex->to_UTF8(name_pos, lgth, u8_ptr);
		return Parser_Interface<W>::model_info->symbol_table->LookupOrInsertReserved();
	}
}

// template <>
// int ParsingEngine< X8_Buffer<EBCDIC>, UTF_8 >::Parse_Name() {
// 	int name_pos = AbsPos();
// 	ScanTo(NameFollow);
// 	int lgth = AbsPos()-name_pos;
// //	int nameID = local_EBCDIC_table->Lookup_or_Insert(GetCodeUnitPtr(name_pos), lgth);
// //	if (nameID != 0) return nameID;
// //	else {
// 		int u8_lgth = byteplex->UTF8_Length(name_pos, lgth);
// 		char * u8_ptr = Parser_Interface<UTF_8>::model_info->symbol_table->ReserveSymbolSpace(u8_lgth);
// 		byteplex->to_UTF8(name_pos, lgth, u8_ptr);
// 		return Parser_Interface<UTF_8>::model_info->symbol_table->LookupOrInsertReserved();
// //	}
// }

// template <WorkingCharacterSet W>
// inline int ParsingEngine<UTF8_Buffer, W>::Parse_Name() {
// 	int name_pos = AbsPos();
// 	ScanTo(NameFollow);
// 	int lgth = AbsPos()-name_pos;
// 	return Parser_Interface<UTF_8>::model_info->symbol_table->UTF8_Lookup_or_Insert_Name(&((char *)x8data)[buffer_rel_pos-lgth], lgth);
// }

template <>
inline int ParsingEngine<UTF8_Buffer, UTF_8>::Parse_Name() {
	int name_pos = AbsPos();
	ScanTo(NameFollow);
	int lgth = AbsPos()-name_pos;
	return Parser_Interface<UTF_8>::model_info->symbol_table->UTF8_Lookup_or_Insert_Name(&((char *)x8data)[buffer_rel_pos-lgth], lgth);
}

template <class B, WorkingCharacterSet W>
int ParsingEngine<B, W>::Parse_Nmtoken() {
	int name_pos = AbsPos();
	ScanTo(NameFollow);
	int lgth = AbsPos()-name_pos;
	int nameID = Parser_Interface<W>::model_info->symbol_table->ASCII_Lookup_or_Insert_Nmtoken(&((char *) x8data)[buffer_rel_pos-lgth], lgth);
	if (nameID != 0) return nameID;
	else {
		int u8_lgth = byteplex->UTF8_Length(name_pos, lgth);
		char * u8_ptr = Parser_Interface<W>::model_info->symbol_table->ReserveSymbolSpace(u8_lgth);
		byteplex->to_UTF8(name_pos, lgth, u8_ptr);
		return Parser_Interface<W>::model_info->symbol_table->LookupOrInsertReserved_nmtoken();
	}
}

/*template <>
int ParsingEngine< X8_Buffer<EBCDIC>, UTF_8 >::Parse_Nmtoken() {
	int name_pos = AbsPos();
	ScanTo(NameFollow);
	int lgth = AbsPos()-name_pos;
//	int nameID = local_EBCDIC_table->Lookup_or_Insert(GetCodeUnitPtr(name_pos), lgth);
//	if (nameID != 0) return nameID;
//	else {
		int u8_lgth = byteplex->UTF8_Length(name_pos, lgth);
		char * u8_ptr = Parser_Interface<UTF_8>::model_info->symbol_table->ReserveSymbolSpace(u8_lgth);
		byteplex->to_UTF8(name_pos, lgth, u8_ptr);
		return Parser_Interface<UTF_8>::model_info->symbol_table->LookupOrInsertReserved_nmtoken();
//	}
}*/
// template <>
// int ParsingEngine<UTF8_Buffer, UTF_8>::Parse_Nmtoken() {
// 	int name_pos = AbsPos();
// 	ScanTo(NameFollow);
// 	int lgth = AbsPos()-name_pos;
// 	return Parser_Interface<UTF_8>::model_info->symbol_table->UTF8_Lookup_or_Insert_Nmtoken(&((char *)x8data)[buffer_rel_pos-lgth], lgth);
// }

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::Parse_DocumentContent() {
	Parser_Interface<W>::DocumentStart_action();
#if (VALIDATION_MODE == ON)
	int cur_state = 0;
	Parse_ValidContent(Parser_Interface<W>::model_info->rootModel, cur_state);
	if (Parser_Interface<W>::model_info->rootModel->transition_map[cur_state][0]==0) {
		Validity_Error(vErr_elementvalid);
	}
#endif
#if (VALIDATION_MODE == OFF)
	Parse_WF_Element();
	ScanTo(NonWS);
	while(at_Comment_Start<B::Base>(cur()) || at_PI_Start<B::Base>(cur()) ){
		if (at_Comment_Start<B::Base>(cur()))
			Parse_Comment();
		else
			Parse_PI();
		ScanTo(NonWS);
	}
	if (!at_EOF()) {
		Syntax_Error(NT_document);
	}
#endif
	Parser_Interface<W>::DocumentEnd_action();
}

#ifdef MARKUP_PASS_CONTROL
// Test routine as an alternative to MarkupPass.
template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::ParseContent() {
	int start_code = 0;
	int end_code = 0;
	int charref_code = 0;
	int general_ref_code = 0;
	DocumentStart_action();
	bool is_emptyStartTag = false;
	do {
		text_or_markup_start = AbsPos();
		ScanTo(MarkupStart); /* '<', '&', or ']' for 0b11']]>' test */
/*		if (AtChar<B::Base,'<'>(cur())) {
			text_if_nonnull_action();
			Parse_Markup<B, W>();
		}*/
		if (at_EndTag_Start<B::Base>(cur())) {
			end_code |= AbsPos();
		}
		else if (AtChar<B::Base,'<'>(cur())) {
			start_code += AbsPos();
		}
		else if (at_CharRef_Start<B::Base>(cur())) {
			charref_code += 1;
		}
		else  if (AtChar<B::Base,'&'>(cur())) {
			general_ref_code += 1;
		}
		else if (at_EOF()) break;
		Advance(1);
	} while (1);
	printf("Start_code: %i\n", start_code);
	printf("End_code: %i\n", end_code);
	printf("general_ref_code: %i\n", general_ref_code);
	printf("charref_code: %i\n", charref_code);
	DocumentEnd_action();
}
#endif

#ifdef MARKUP_SORTING
// Little endian codes for [&#/] stream.
enum MarkupSortCodes {
  StartTagTwoBitCode = 0,
  EndTagTwoBitCode = 2,
  GeneralRefCode = 1,
  CharRefCode = 3
};


static inline int GetBitPair(SIMD_type * stream, int bit_posn) {
	return bitstream_segment_from(stream, bit_posn) & 3;
}

template <class B, WorkingCharacterSet W>
void ParsingEngine<B, W>::ParseContent() {
/*vector<int> MarkupPositions[4];*/
int MarkupPositions[4][BUFFER_SIZE];
int MarkupCounts[4];
	int start_code = 0;
	int end_code = 0;
	int charref_code = 0;
	int general_ref_code = 0;

	DocumentStart_action();
	bool is_emptyStartTag = false;
		for (int i = 0; i < 4; i++) MarkupCounts[i] = 0;
		text_or_markup_start = AbsPos();
	do {
		unsigned long segment = bitstream_segment_from(buf->item_stream[MarkupStart], buffer_rel_pos);
//printf("buffer_rel_pos = %i, segment = %x\n", buffer_rel_pos, segment);
		if (segment != 0) {
			buffer_rel_pos += cfzl(segment);
		text_or_markup_start = AbsPos();
			int markup_code = GetBitPair(buf->item_stream[AmpHashSlash], buffer_rel_pos);
			MarkupPositions[markup_code][MarkupCounts[markup_code]] = AbsPos();
			MarkupCounts[markup_code]++;
			Advance(1);
		}
		else {
			buffer_rel_pos = (buffer_rel_pos & -8) + 8*sizeof(unsigned long);
// printf("buffer_rel_pos = %i, segment = %x\n", buffer_rel_pos, segment);

			if (buffer_rel_pos >= buffer_limit_pos) {
/*				for (int i = 0; i < MarkupCounts[StartTagTwoBitCode]; i++) {
					start_code += MarkupPositions[StartTagTwoBitCode][i];
				}
				for (int i = 0; i < MarkupCounts[EndTagTwoBitCode]; i++) {
					end_code |= MarkupPositions[EndTagTwoBitCode][i];
				}
				for (int i = 0; i < MarkupCounts[GeneralRefCode]; i++) {
					general_ref_code += 1;
				}
				for (int i = 0; i < MarkupCounts[CharRefCode]; i++) {
					charref_code += 1;
				}*/
/*	printf("Start_code: %i\n", start_code);
	printf("End_code: %i\n", end_code);
	printf("general_ref_code: %i\n", general_ref_code);
	printf("charref_code: %i\n", charref_code);*/
				for (int i = 0; i < 4; i++) MarkupCounts[i] = 0;
				if (buffer_rel_pos >= BUFFER_SIZE) {
					AdjustBufferEndForIncompleteSequences();
					Parser_Interface<W>::FinalizeBuffer_action();
					AdvanceBuffers();
				}
				else break;
			}

		}

 	} while (1);
/*	vector<int>::iterator i;
	for (i = MarkupPositions[StartTagTwoBitCode].begin(); i != MarkupPositions[StartTagTwoBitCode].end(); i++) {
		start_code += *i;
	}
	for (i = MarkupPositions[EndTagTwoBitCode].begin(); i != MarkupPositions[EndTagTwoBitCode].end(); i++) {
		end_code |= *i;
	}
	for (i = MarkupPositions[GeneralRefCode].begin(); i != MarkupPositions[GeneralRefCode].end(); i++) {
		general_ref_code += 1;
	}
	for (i = MarkupPositions[CharRefCode].begin(); i != MarkupPositions[CharRefCode].end(); i++) {
		charref_code += 1;
	}*/
	printf("Start_code: %i\n", start_code);
	printf("End_code: %i\n", end_code);
	printf("general_ref_code: %i\n", general_ref_code);
	printf("charref_code: %i\n", charref_code);
	DocumentEnd_action();
}

#endif




